#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>

#include "adc.h"
#include "serial.h"
#include "macros.h"
#include "hbridge.h"

// Uncomment one of these to switch minutes/seconds mode.
#define MINUTES
//#define HOURS

// Uncomment these flags to change behavior
//#define DEMO // Run 60x faster
#define TALKATIVE // Output status on serial line

volatile unsigned char seconds = 0;
#ifdef HOURS
volatile unsigned char minutes = 0;
#endif

volatile unsigned char increment_now = 0;

// 61 roughly equidistant locations along the slider
#ifdef MINUTES
int calibration[] PROGMEM =
{
41, 55, 66, 81, 93,
105, 121, 130, 143, 154,
171, 189, 200, 218, 232,
254, 288, 307, 339, 400,
472, 562, 612, 655, 718,
850, 911, 961, 1007, 1067,
1157, 1235, 1250, 1260, 1265,
1287, 1298, 1312, 1331, 1343,
1358, 1378, 1394, 1413, 1444,
1505, 1550, 1588, 1631, 1657,
1704, 1737, 1769, 1813, 1848,
1902, 1937, 1971, 2013, 2030,
2046
};
#endif

#ifdef HOURS
int calibration[] PROGMEM =
{
40, 47, 56, 64, 74,
89, 95, 104, 112, 123,
137, 148, 161, 174, 186,
211, 240, 265, 295, 340,
445, 520, 575, 627, 659,
771, 840, 890, 960, 1050,
1122, 1200, 1242, 1257, 1266,
1285, 1300, 1313, 1327, 1339,
1362, 1379, 1395, 1420, 1456,
1503, 1545, 1580, 1615, 1660,
1707, 1748, 1794, 1830, 1855,
1904, 1948, 1980, 2006, 2038,
2046
};
#endif

// There are 60 unique slider positions.  Minutes go from 0 - 59 in
// one minute intervals, hours 1:00 to 12:44 in 16 minute increments
// (so 5 increments per hour).

// There are 61 calibration positions so that the slider is left
// between a pair of calibration positions.  This ensures that we
// don't accidentally skip or repeat positions.

static const int calibration_count = 61;

// External header (top to bottom):
// GND
// D3, 4 (from bridge)
// D1, 2 (from bridge)
// ADC6
// PB5 (input voltage to potentiometer)
    

void timer2_init(void)
{
    clear(ASSR, EXCLK); // Disable synchronous clock  
    set(ASSR, AS2); // Enable asynchronous timer crystal

    // Timer 2 mode = normal
    TCCR2A = 0;
    // Timer 2 prescalar = 128
    TCCR2B = 0b101;
    
    set(TIMSK2, TOIE2); // Enable Timer2 interrupts
}

void go_to(unsigned int target)
{
    if (target < 0 || target > 2046)
        return;
    unsigned char speed;
    int a = adc_read_combined();
    for(speed = 200; speed >= 40; speed -= 40) {      
        if (a < target) {
            motor_forward_pwm(speed);
            while(a < target)
                a = adc_read_combined();
        } else if (a > target) {
            motor_backward_pwm(speed);
            while(a > target)
                a = adc_read_combined();
        } else {
            motor_stop();
            return; // If we've actually hit the target.
        }
        motor_stop();
        _delay_ms(100);
        a = adc_read_combined();
    }
    motor_stop();
}

void go_short(unsigned int target)
{
    if (target < 0 || target > 2046)
        return;
        
    int a = adc_read_combined();
    
    if (a < target) {
        motor_forward_pwm(40);
        while(adc_read_combined() < target);
    } else if (a > target) {
        motor_backward_pwm(40);
        while(adc_read_combined() > target);
    } else {
        return; // If we've actually hit the target.
    }
    
    motor_stop();
}

void calibrate(void)
{
    #ifdef TALKATIVE
    // Collect calibration data
    static char message[] PROGMEM = "Calibration data:\n";
    put_flash_string((PGM_P) message);
    
    go_to(44);
    _delay_ms(1000);
    int a = 0;
    int i = 0;
    while(a < 2043) {
        _delay_ms(58); // Time it so we get about 62 data points
//        motor_stop();
//        _delay_ms(250);
        a = adc_read_combined();
        put_int(a);
        put_char('\n');
        motor_forward_pwm(50);
        i++;
    }
    motor_stop();
    put_char('\n');
    static char m2[] PROGMEM = " data points collected.\n";
    put_int(i);
    put_flash_string((PGM_P) m2);
    #endif
}

// An interactive pseudo-terminal.
void interactive(void)
{
    char c;
    
    while(1) {
        while(in_buffer_empty()) {
                put_int(adc_read_combined());
                put_char('\n');
        _delay_ms(100);}
        
        // 'g #' for go to raw location.
        // 'r' for read 
        // 'c' for calibrate
        get_char(&c);
        put_char(c);
        if (c == 'g') {
            int loc = 0;
            // extract a number from the string
            while(c != 10) {
                while(in_buffer_empty()); // Wait for a character
                get_char(&c);
                put_char('c');
                put_char(c);
                if (c >= '0' && c <= '9')
                    loc = loc * 10 + c - '0';
            }
            put_int(loc);
            put_char('\n');
            go_to(loc);
        } else if (c == 'r') {
            put_int(adc_read_combined());
            put_char('\n');
        } else if (c == 'c')
            calibrate();
    }
}

void power_down(void)
{   
    clear(ADCSRA, ADEN); // disable ADC
    set(PRR, PRADC); // turn off power to ADC
    
    set(PRR, PRTIM0); // turn off power to timer0
}

void power_up(void)
{
    clear(PRR, PRADC);
    set(ADCSRA, ADEN);
    clear(PRR, PRTIM0);
}
    

int main(void)
{
    //
    // main
    //
    
    // Initialize various special features
    adc_init();
    hbridge_init();
    serial_init();
    timer2_init();

    // Initialize the LED as an output
    set(DDRD, PD2);

    // Flash the LED, give Timer2 a chance to start up.
    unsigned char i;
    for (i = 0; i < 20; ++i) {
        set(PORTD, PD2);
        _delay_ms(25);
        clear(PORTD, PD2);
        _delay_ms(25);
    }    
    
    // Turn off power to anything not be used.
    set(PRR, PRTWI); // turn off two-wire interface
    set(PRR, PRSPI); // turn off SPI
    #ifndef TALKATIVE
        set(PRR, PRUSART0); // turn off USART
    #endif
    set(ACSR, ACD); // turn off analog comparator

    // Enable global interrupts (for timer things)
    sei();
    
//    interactive();
    put_char('\n');
    
    while(1) {
        power_down();
        set_sleep_mode(SLEEP_MODE_PWR_SAVE);
        sleep_enable();
        sleep_cpu();
        sleep_disable();
        
        // Clear the LED heartbeat.
        if (bit_test(PORTD, PD2)) {
            _delay_ms(100);
            clear(PORTD, PD2);
        }

        if (increment_now) {
            power_up();
            _delay_ms(250);
            increment_now = 0;
            char longjump = 0;
            int loc = adc_read_combined();
            
            #ifdef TALKATIVE
                put_char('\n');
                put_char('\n');
                put_char('L');
                put_char(' ');
                put_int(loc);
                put_char('\n');
            #endif
            
            int i;
            // Figure out where we are on the slider.
            for (i = 1; i < calibration_count; ++i) {
                if (loc < pgm_read_word(&calibration[i]))
                    break;
            }
            // If we're in the topmost region (or above it!), then
            // reset the slider.
            if (i >= calibration_count - 1) {
                i = 0;
                longjump = 1;
            }
            
            // Figure out the new slider position.  It will be in
            // between a pair of calibration points.
            int lower = pgm_read_word(&calibration[i++]);
            int upper = pgm_read_word(&calibration[i++]);
            
            int newloc = (lower + upper) >> 1;
            
            #ifdef TALKATIVE
                put_char('i');
                put_char(' ');
                put_int(i);
                put_char('\n');
            
                put_char('N');
                put_char(' ');
                put_int(newloc);
                put_char('\n');
            #endif
            
            // We use a fast jump if we need to get all the way back
            // to the beginning of the slide.
            if (longjump) {
                go_to(newloc);
                _delay_ms(250);
                loc = adc_read_combined();
            }
                
            // Loop until we get where we want to be.
            i = 0;
            while(loc <= lower || loc >= upper) {
                // Occasionally, we get stuck bouncing between two points.
                // This encourages a big move in that situation.
                if (!(++i % 10))
                    go_to(newloc);
                else
                    go_short(newloc);
                _delay_ms(250);
                loc = adc_read_combined();
            }
                
        }
    }

	 return 0;
}

ISR(TIMER2_OVF_vect) {

    if (!(seconds%5))
        set(PORTD, PD2); // Heartbeat at five second intervals.

    #ifdef DEMO
    if (++seconds == 2) {
    #else
    if (++seconds == 60) {
    #endif
    
        seconds = 0;
        
        // For minutes, increment the slider every minute.
        # ifdef MINUTES
            increment_now = 1;
        // For hours, increment the slider every 12 minutes (5x/hour)
        // If we're in demo mode, increment the hours slider once per minute.
        # else
            if (++minutes == 12) {
                minutes = 0;
                increment_now = 1;
            }
        #endif
    }
}